#include-once

#cs
This set of functions is intended to handle GUI messages flow with advantages: 
+ messages for different GUI can have independent handlers (messages registered with GUIRegisterMsg are script-wide)
+ easily variables exchange between caller and callee (message handler) functions without additional globals

; But this has disadvantage - slightly slower (especially when many handlers registered) then directly
; calling functions. This will be be rewritten with maps for speed if maps will be finally released.
; I know about Scripting.Dictionary, but not want rely on it.
#ce

; ==================================================================================================
; 2021-09-08 - Fixed unregister code, redone some logic, simplified and optimized code
; 2021-09-04 - Moved to separate UDF MSGHandler.au3
; 2017-08-17 - Initial release
; ==================================================================================================

Global Enum $MH_MSG, $MH_FUNC, $MH_DATA, $MH_HWND, $MH_ACTIVE, $MH_MAX
Global Enum $MH_DBKCLICK_CTRL, $MH_DBKCLICK_INVOKE

Global $g_aGlobalMessageHandlerList[0][$MH_MAX] ; UBound is about 15% faster then retrieve zero-item

Global Const $_NH_WM_COMMAND = 0x0111
Global Const $_NH_WM_NOTIFY  = 0x004E
Global Const $_NH_LBN_DBLCLK = 0x00000002
Global Const $_NH_NM_DBLCLK  = -3

Global Const $_GUI_RUNDEFMSG = 'GUI_RUNDEFMSG'

; ==================================================================================================
; Register new message
;
; If $vData = Default, standard 4-params message function used
; When assigned data, 5-th param of function is data passed
; ==================================================================================================
Func _MSGHandler_Register($hWnd, $nMessage, $fFunc, $vData = Default)
    ; find element with "free" state
    Local $nIndex = -1
    For $i = 0 To UBound($g_aGlobalMessageHandlerList) - 1
        If $g_aGlobalMessageHandlerList[$i][$MH_ACTIVE] = False Then
            $nIndex = $i
            ExitLoop
        EndIf
    Next
    ; if not found, add array element
    If $nIndex = -1 Then
        $nIndex = UBound($g_aGlobalMessageHandlerList)
        ReDim $g_aGlobalMessageHandlerList[$nIndex + 1][$MH_MAX]
    EndIf

    $g_aGlobalMessageHandlerList[$nIndex][$MH_HWND]   = $hWnd
    $g_aGlobalMessageHandlerList[$nIndex][$MH_MSG]    = $nMessage
    $g_aGlobalMessageHandlerList[$nIndex][$MH_FUNC]   = $fFunc
    $g_aGlobalMessageHandlerList[$nIndex][$MH_DATA]   = $vData
    $g_aGlobalMessageHandlerList[$nIndex][$MH_ACTIVE] = True

    GUIRegisterMsg($g_aGlobalMessageHandlerList[$nIndex][$MH_MSG], "__MessageHandlerDispatcher")

    Return $nIndex
EndFunc

; ==================================================================================================
; Unregister message 
; 
; It's important to unregister messages when GUI deleted, else it's will slow down script after each 
; sub-GUI opening. 
;
; Single message can be unregistered when passing a "handle" returned by _MSGHandler_Register. 
; ==================================================================================================
Func _MSGHandler_Unregister($hWnd, $nIndex = -1)
    Local $aUnreg[0] ; lookup array

    If $nIndex = -1 Then
        ; collect messages for given hWnd
        For $i = 0 To UBound($g_aGlobalMessageHandlerList)-1
            If $g_aGlobalMessageHandlerList[$i][$MH_HWND] = $hWnd Then
                $g_aGlobalMessageHandlerList[$i][$MH_ACTIVE] = False
                ReDim $aUnreg[UBound($aUnreg) + 1]
                $aUnreg[UBound($aUnreg) - 1] = $g_aGlobalMessageHandlerList[$i][$MH_MSG]
            EndIf
        Next
    Else
        ; add single message
        If $nIndex >= 0 And $nIndex < UBound($g_aGlobalMessageHandlerList) Then
            $g_aGlobalMessageHandlerList[$nIndex][$MH_ACTIVE] = False
            ReDim $aUnreg[1]
            $aUnreg[0] = $g_aGlobalMessageHandlerList[$nIndex][$MH_MSG]
        EndIf
    EndIf

    ; find unused messages (would be simpler with Maps)
    For $i = 0 To UBound($g_aGlobalMessageHandlerList)-1
        If Not $g_aGlobalMessageHandlerList[$i][$MH_ACTIVE] Then ContinueLoop
        For $j = 0 To UBound($aUnreg)-1
            If $aUnreg[$j] = $g_aGlobalMessageHandlerList[$i][$MH_MSG] Then $aUnreg[$j] = 0
        Next
    Next

    ; unregister unused messages
    For $nMsg In $aUnreg
        If $nMsg <> 0 Then GUIRegisterMsg($nMsg, "")
    Next
EndFunc

; ==================================================================================================
; Helper to create new params array inline (up to 9 params)
; ==================================================================================================
Func _MSGHandler_MakeData($vParam1 = Default, $vParam2 = Default, $vParam3 = Default, _ 
                          $vParam4 = Default, $vParam5 = Default, $vParam6 = Default, _
                          $vParam7 = Default, $vParam8 = Default, $vParam9 = Default)
    Local $aData[9] = [$vParam1, $vParam2, $vParam3, $vParam4, $vParam5, $vParam6, $vParam7, $vParam8, $vParam9]
    ReDim $aData[@NumParams]
    Return $aData
EndFunc

; ==================================================================================================
; One of most frequent handlers shortcut - doubleclick handler for ListView and ListBox
; 
; CAUTION! If parent GUI use own WM_NOTIFY or WM_COMMAND message handlers, this will overwrite it, so
; it need re-register parent GUI handlers after child closed. Better to use this framework for whole
; script message handlers if possible. 
;
; $hWnd     - Window, where ListView or ListBox created
; $hControl - ListView or ListBox control for which doubleclick will be registered
; $hInvoke  - WM_COMMAND message will be send as from this control in case of doubleclick
; ==================================================================================================
Func _MSGHandler_RegisterDoubleClick($hWnd, $hControl, $hInvoke)
    Local $aData[] = [$hControl, $hInvoke]
    If Not IsHWnd($hControl) Then $hControl = GUICtrlGetHandle($hControl)
    Local $aCall = DllCall("user32.dll", "int", "GetClassNameW", "hwnd", $hControl, "wstr", "", "int", 1024)

    Switch $aCall[2]
        Case "SysListView32"
            Return _MSGHandler_Register($hWnd, $_NH_WM_NOTIFY,  __ListViewDoubleClickFunc, $aData)
        Case "ListBox"
            Return _MSGHandler_Register($hWnd, $_NH_WM_COMMAND, __ListBoxDoubleClickFunc,  $aData)
    EndSwitch
EndFunc

; ==================================================================================================
; Internal use functions
; ==================================================================================================

; ==================================================================================================
; This function dispatch notifications of all types
; ==================================================================================================
Func __MessageHandlerDispatcher($hWnd, $MsgID, $wParam, $lParam)
    Local $vResult, $vData
    For $vKey = 0 To UBound($g_aGlobalMessageHandlerList)-1
        ; don't not run function for windows not match criteria
        If Not $g_aGlobalMessageHandlerList[$vKey][$MH_ACTIVE] Then ContinueLoop
        If $g_aGlobalMessageHandlerList[$vKey][$MH_MSG]  <> $MsgID Then ContinueLoop
        If $g_aGlobalMessageHandlerList[$vKey][$MH_HWND] <> $hWnd Then ContinueLoop

        ; execute assigned function
        $vData = $g_aGlobalMessageHandlerList[$vKey][$MH_DATA]
        If $vData = Default Then
            $vResult = ($g_aGlobalMessageHandlerList[$vKey][$MH_FUNC])($hWnd, $MsgID, $wParam, $lParam)
        Else
            $vResult = ($g_aGlobalMessageHandlerList[$vKey][$MH_FUNC])($hWnd, $MsgID, $wParam, $lParam, $vData)
        EndIf
        ; we can have few messages of same type (for different controls), so check if message handled in this call 
        ; if handled, return result, if not - continue loop to check next control
        if $vResult <> $_GUI_RUNDEFMSG Then Return $vResult
    Next
    Return $_GUI_RUNDEFMSG ; nothing handled
EndFunc

Func __ListViewDoubleClickFunc($hWnd, $MsgID, $wParam, $lParam, $vData)
    Local $tNMHDR = DllStructCreate("hwnd hwndFrom;uint idFrom;uint code", $lParam)

    Local $nCode = Int($tNMHDR.code, 1) ; it's important to interpret code as 32-bit integer
    Local $nIDFrom = BitAND($wParam, 0xFFFF)

    If ($nIDFrom = $vData[$MH_DBKCLICK_CTRL]) And ($nCode = $_NH_NM_DBLCLK) Then
        DllCall("user32.dll", "int", "PostMessage", "hwnd", $hWnd, "int", $_NH_WM_COMMAND, "wparam", $vData[$MH_DBKCLICK_INVOKE], "lparam", GUICtrlGetHandle($vData[$MH_DBKCLICK_INVOKE]))
        Return ; handled
    EndIf 
    Return $_GUI_RUNDEFMSG
EndFunc

Func __ListBoxDoubleClickFunc($hWnd, $MsgID, $wParam, $lParam, $vData)
    Local $nCode   = BitShift($wParam, 16)
    Local $nIDFrom = BitAND($wParam, 0xFFFF)

    If ($nIDFrom = $vData[$MH_DBKCLICK_CTRL]) And ($nCode = $_NH_LBN_DBLCLK) Then
        DllCall("user32.dll", "int", "PostMessage", "hwnd", $hWnd, "int", $_NH_WM_COMMAND, "wparam", $vData[$MH_DBKCLICK_INVOKE], "lparam", GUICtrlGetHandle($vData[$MH_DBKCLICK_INVOKE]))
        Return ; handled
    EndIf 
    Return $_GUI_RUNDEFMSG
EndFunc
